#! /usr/bin/env python
# encoding: utf-8
#
# This file is part of MXCuBE.
#
# MXCuBE is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# MXCuBE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with MXCuBE.  If not, see <https://www.gnu.org/licenses/>.
"""Xray Centring Abstract Class with yaml configuration file."""

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

__copyright__ = """ Copyright © 2016 - 2022 by MXCuBE Collaboration """
__license__ = "LGPLv3+"
__author__ = "rhfogh"
__date__ = "25/03/2022"

import abc

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.model import queue_model_objects


class AbstractXrayCentring(HardwareObject):
    """Xray Centring Hardware Object. Set to Yaml configuration."""

    def __init__(self, name):
        super().__init__(name)

        # Needed to allow methods to put new actions on the queue
        # And as a place to get hold of other objects, like the QMO
        self._queue_entry = None

        # Current data collection task group. For adding tasks to queue. if needed
        self._data_collection_group = None

    def _init(self):
        """This function is superfluous, but left in as a reminder in
        case you need to override it.
        """
        super()._init()

    def init(self):
        """Initialise the state to READY"""
        super().init()
        self.update_state(self.STATES.READY)

    def shutdown(self):
        """Shut down Xray centring. Triggered on program quit."""

    def pre_execute(self, queue_entry):
        """Actions to take before the execution of the centring.
        Args:
            queue_entry(XrayCentring2QueueEntry): Access to queue and data model
        """

        self._queue_entry = queue_entry
        if self.is_ready():
            self.update_state(self.STATES.BUSY)
        else:
            raise RuntimeError(
                "Cannot execute Xray centring - HardwareObject is not ready"
            )
        new_dcg_model = queue_model_objects.TaskGroup()
        new_dcg_model.set_enabled(True)
        new_dcg_model.set_name("CentringTaskGroup")
        self._data_collection_group = new_dcg_model
        self._add_to_queue(self._queue_entry.get_data_model(), new_dcg_model)

    @abc.abstractmethod
    def execute(self):
        """Contains the actual X-ray centring code
        taking parameters from the data model as needed.
        """

    def post_execute(self):
        """The workflow has finished, sets the state to 'READY'."""

        self._queue_entry = None
        self._data_collection_group = None
        self.update_state(self.STATES.READY)

    def _add_to_queue(self, parent_model_obj, child_model_obj):
        """Add entries to queue while centring process is running (if needed).
        Args:
            parent_model_obj: Parent queue entry.
            child_model_obj: Child queue entry.
        """
        HWR.beamline.queue_model.add_child(parent_model_obj, child_model_obj)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""EMBLBSD (Beam shaping device) represents a diffractometer without a gonio"""

import logging
import time

import gevent

from mxcubecore.HardwareObjects.GenericDiffractometer import GenericDiffractometer

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLBSD(GenericDiffractometer):
    """EMBLBSD represents a diffractometer without a gonio"""

    def __init__(self, *args):
        """Based on the GenericDiffractometer without centering motors"""

        GenericDiffractometer.__init__(self, *args)

        # Hardware objects ----------------------------------------------------
        self.zoom_motor_hwobj = None
        self.omega_reference_motor = None

        # Channels and commands -----------------------------------------------
        self.chan_beamstop_position = None
        self.chan_calib_x = None
        self.chan_calib_y = None
        self.chan_current_phase = None
        self.chan_fast_shutter_is_open = None
        self.chan_state = None
        self.chan_sync_move_motors = None
        self.chan_scintillator_position = None
        self.chan_capillary_position = None
        self.chan_status = None
        self.cmd_start_set_phase = None
        self.cmd_start_auto_focus = None
        self.cmd_save_centring_positions = None

        # Internal values -----------------------------------------------------
        self.use_sc = False
        self.omega_reference_par = None
        self.omega_reference_pos = [0, 0]
        self.current_state = None
        self.fast_shutter_is_open = None

    def init(self):
        """Initiates variables"""

        GenericDiffractometer.init(self)

        self.chan_state = self.get_channel_object("State")
        self.current_state = self.chan_state.get_value()
        self.chan_state.connect_signal("update", self.state_changed)

        self.chan_status = self.get_channel_object("Status")
        self.chan_status.connect_signal("update", self.status_changed)

        self.chan_calib_x = self.get_channel_object("CoaxCamScaleX")
        self.chan_calib_y = self.get_channel_object("CoaxCamScaleY")
        self.update_pixels_per_mm()

        self.chan_current_phase = self.get_channel_object("CurrentPhase")
        self.connect(self.chan_current_phase, "update", self.current_phase_changed)

        self.chan_fast_shutter_is_open = self.get_channel_object("FastShutterIsOpen")
        self.chan_fast_shutter_is_open.connect_signal(
            "update", self.fast_shutter_state_changed
        )

        self.chan_scintillator_position = self.get_channel_object(
            "ScintillatorPosition"
        )
        self.chan_capillary_position = self.get_channel_object("CapillaryPosition")

        self.cmd_start_set_phase = self.get_command_object("startSetPhase")
        self.cmd_start_auto_focus = self.get_command_object("startAutoFocus")

        self.zoom_motor_hwobj = self.get_object_by_role("zoom")
        self.connect(self.zoom_motor_hwobj, "valueChanged", self.zoom_position_changed)
        self.connect(
            self.zoom_motor_hwobj,
            "predefinedPositionChanged",
            self.zoom_motor_predefined_position_changed,
        )

        self.chan_beamstop_position = self.get_channel_object("BeamstopPosition")

    def use_sample_changer(self):
        """Returns true if sample changer is used

        :return: bool
        """
        return False

    def manual_centring(self):
        """No need to implement"""
        return

    def automatic_centring(self):
        """No need to implement"""
        return

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        """No need to implement"""
        return

    def state_changed(self, state):
        """Emits state change signal"""
        self.current_state = state
        self.emit("minidiffStateChanged", (self.current_state))
        self.emit("minidiffStatusChanged", (self.current_state))

    def status_changed(self, state):
        """Emits status message"""
        self.emit("statusMessage", ("diffractometer", state, "busy"))

    def zoom_position_changed(self, value):
        """After the zoom change updates pixels per mm"""
        self.update_pixels_per_mm()
        self.current_motor_positions["zoom"] = value

    def zoom_motor_predefined_position_changed(self, position_name, offset):
        """Updates pixels per mm when the zoom has been changed"""
        self.update_pixels_per_mm()
        self.emit("zoomMotorPredefinedPositionChanged", (position_name, offset))

    def fast_shutter_state_changed(self, is_open):
        """Stores fast shutter state"""
        self.fast_shutter_is_open = is_open
        if is_open:
            msg = "Opened"
        else:
            msg = "Closed"
        self.emit("minidiffShutterStateChanged", (self.fast_shutter_is_open, msg))

    def update_pixels_per_mm(self, *args):
        """Updates pixels per mm"""
        if self.chan_calib_x:
            self.pixels_per_mm_x = 1.0 / self.chan_calib_x.get_value()
            self.pixels_per_mm_y = 1.0 / self.chan_calib_y.get_value()
            self.emit(
                "pixelsPerMmChanged", ((self.pixels_per_mm_x, self.pixels_per_mm_y),)
            )

    def set_phase(self, phase, timeout=80):
        """Sets diffractometer to the selected phase.
           In the plate mode before going to or away from
           Transfer or Beam location phase if needed then detector
           is moved to the safe distance to avoid collision.
        """
        # self.wait_device_ready(2)
        logging.getLogger("GUI").warning(
            "Diffractometer: Setting %s phase. Please wait..." % phase
        )

        if timeout is not None:
            _start = time.time()
            self.cmd_start_set_phase(phase)
            gevent.sleep(5)
            with gevent.Timeout(
                timeout, Exception("Timeout waiting for phase %s" % phase)
            ):
                while phase != self.chan_current_phase.get_value():
                    gevent.sleep(0.1)
            self.wait_device_ready(30)
            self.wait_device_ready(30)
            _howlong = time.time() - _start
            if _howlong > 11.0:
                logging.getLogger("GUI").error(
                    "Changing phase to %s took %.1f seconds" % (phase, _howlong)
                )
        else:
            self.cmd_start_set_phase(phase)

    def start_auto_focus(self, timeout=None):
        """Autofocus method"""
        if timeout:
            self.ready_event.clear()
            gevent.spawn(self.execute_server_task, self.cmd_start_auto_focus(), timeout)
            self.ready_event.wait()
            self.ready_event.clear()
        else:
            self.cmd_start_auto_focus()

    def emit_diffractometer_moved(self, *args):
        """Emits diffractometerMoved signal"""
        self.emit("diffractometerMoved", ())

    def re_emit_values(self):
        """Reemits all signals"""
        self.emit("minidiffPhaseChanged", (self.current_phase,))
        self.emit("minidiffShutterStateChanged", (self.fast_shutter_is_open,))
        self.emit("pixelsPerMmChanged", ((self.pixels_per_mm_x, self.pixels_per_mm_y),))

    def move_omega(self, angle):
        """No need to implement"""
        return

    def set_zoom(self, position):
        """Sets zoom"""
        self.zoom_motor_hwobj.move_to_position(position)

    def get_osc_limits(self):
        """Return oscillation limits"""
        return (-1e6, 1e6)

    def get_scintillator_position(self):
        """Returns scintillator position"""
        return self.chan_scintillator_position.get_value()

    def set_scintillator_position(self, position):
        """Sets scintillator position"""
        self.chan_scintillator_position.set_value(position)
        with gevent.Timeout(5, Exception("Timeout waiting for scintillator position")):
            while position != self.get_scintillator_position():
                gevent.sleep(0.01)

    def get_capillary_position(self):
        """Returns capillary position"""
        return self.chan_capillary_position.get_value()

    def set_capillary_position(self, position):
        """Moves capillary to requested position"""
        self.chan_capillary_position.set_value(position)
        with gevent.Timeout(5, Exception("Timeout waiting for capillary position")):
            while position != self.get_capillary_position():
                gevent.sleep(0.01)

    def zoom_in(self):
        """Steps zoom one step in"""
        self.zoom_motor_hwobj.zoom_in()

    def zoom_out(self):
        """Steps zoom one step out"""
        self.zoom_motor_hwobj.zoom_out()

    def set_beamstop_park(self):
        """Sets beamstop to the parking position"""
        self.chan_beamstop_position.set_value("PARK")

    def set_beamstop_beam(self):
        """Sets beamstop in the beam position"""
        self.chan_beamstop_position.set_value("BEAM")



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


"""Sardana Control System"""

from __future__ import absolute_import

import logging
import os

import gevent
from gevent.event import Event

# import time
# import types
from mxcubecore.dispatcher import saferef

try:
    import Queue as queue
except ImportError:
    import queue

gevent_version = list(map(int, gevent.__version__.split(".")))


from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
    ConnectionError,
)

# This is a site specific module where some of the dependencies might not be capture by the ``pyproject.toml`` during installation

try:
    import PyTango
    from PyTango import (
        ConnectionFailed,
        DevFailed,
    )
except Exception:
    logging.getLogger("HWR").warning("Pytango is not available in this computer.")

#

try:
    import taurus
    from sardana.taurus.core.tango.sardana import registerExtensions
    from taurus import (
        Attribute,
        Device,
    )
except Exception:
    logging.getLogger("HWR").warning("Sardana is not available in this computer.")


__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


def processSardanaEvents():
    while not SardanaObject._eventsQueue.empty():
        try:
            ev = SardanaObject._eventsQueue.get_nowait()
        except queue.Empty:
            break
        else:
            try:
                receiver_cb_ref = SardanaObject._eventReceivers[id(ev)]
                receiver_cb = receiver_cb_ref()
                if receiver_cb is not None:
                    try:
                        gevent.spawn(receiver_cb, ev)
                    except AttributeError:
                        pass
            except KeyError:
                pass


def wait_end_of_command(cmdobj):
    while (
        cmdobj.macrostate == SardanaMacro.RUNNING
        or cmdobj.macrostate == SardanaMacro.STARTED
    ):
        gevent.sleep(0.05)
    return cmdobj.door.result


def end_of_macro(macobj):
    macobj._reply_arrived_event.wait()


class AttributeEvent:
    def __init__(self, event):
        self.event = event


class SardanaObject(object):
    """Sardana Object"""

    _eventsQueue = queue.Queue()
    _eventReceivers = {}

    if gevent_version < [1, 3, 0]:
        _eventsProcessingTimer = getattr(gevent.get_hub().loop, "async")()
    else:
        _eventsProcessingTimer = gevent.get_hub().loop.async_()

    # start Sardana events processing timer
    _eventsProcessingTimer.start(processSardanaEvents)

    def object_listener(self, *args):
        ev = AttributeEvent(args)
        # NBNB self.update not defined
        SardanaObject._eventReceivers[id(ev)] = saferef.safe_ref(self.update)
        SardanaObject._eventsQueue.put(ev)
        SardanaObject._eventsProcessingTimer.send()


class SardanaMacro(CommandObject, SardanaObject, ChannelObject):
    """Sardana macro"""

    macroStatusAttr = None
    INIT, STARTED, RUNNING, DONE = range(4)

    def __init__(self, name, macro, doorname=None, username=None, **kwargs):
        super(SardanaMacro, self).__init__(name, username, **kwargs)

        self._reply_arrived_event = Event()
        self.macro_format = macro
        self.doorname = doorname
        self.door = None
        self.id_result = -1
        self.macrostate = SardanaMacro.INIT
        self.doorstate = None
        self.t0 = 0
        self.init_device()

    def init_device(self):
        self.door = Device(self.doorname)
        self.door.set_timeout_millis(10000)
        self.doorstate = self.door.state.name.upper()

        #
        # DIRTY FIX to make compatible taurus listeners and existence of Tango channels/commands
        # as defined in Command/Tango.py
        #
        # if self.door.__class__ == taurus.core.tango.tangodevice.TangoDevice:
        #    dp = self.door.getHWObj()
        #    try:
        #        dp.subscribe_event = dp._subscribe_event
        #    except AttributeError:
        #        pass

        if self.macroStatusAttr is None:
            self.macroStatusAttr = self.door.getAttribute("State")
            self.macroStatusAttr.addListener(self.object_listener)

    def result_callback(self, received_event):
        val = received_event.attr_value.value
        if val is not None:
            self.emit("macroResultUpdated", received_event.attr_value.value)

    def __call__(self, *args, **kwargs):
        self._reply_arrived_event.clear()
        self.result = None

        wait = kwargs.get("wait", False)

        if self.door is None:
            self.init_device()

        logging.getLogger("HWR").debug(
            "Executing sardana macro: %s" % self.macro_format
        )
        logging.getLogger("HWR").debug(
            "   args=%s / kwargs=%s" % (str(args), str(kwargs))
        )

        try:
            fullcmd = self.macro_format + " " + " ".join([str(a) for a in args])
        except Exception:
            import traceback

            logging.getLogger("HWR").info(traceback.format_exc())
            logging.getLogger("HWR").info(
                "  - Wrong format for macro arguments. Macro is %s / args are (%s)"
                % (self.macro_format, str(args))
            )
            return

        try:
            import time

            self.t0 = time.time()
            if self.doorstate in ["ON", "ALARM", "READY"]:
                self.id_result = self.door.subscribe_event(
                    "Result", PyTango.EventType.CHANGE_EVENT, self.result_callback
                )
                self.door.runMacro(fullcmd.split())
                self.macrostate = SardanaMacro.STARTED
                self.emit("commandBeginWaitReply", (str(self.name()),))
            else:
                logging.getLogger("HWR").error(
                    "%s. Cannot execute. Door is not READY", str(self.name())
                )
                self.emit("commandFailed", (-1, self.name()))
        except TypeError:
            logging.getLogger("HWR").error(
                "%s. Cannot properly format macro code. Format is: %s, args are %s",
                str(self.name()),
                self.macro_format,
                str(args),
            )
            self.emit("commandFailed", (-1, self.name()))
        except DevFailed as error_dict:
            logging.getLogger("HWR").error(
                "%s: Cannot run macro. %s", str(self.name()), error_dict
            )
            self.emit("commandFailed", (-1, self.name()))
        except AttributeError as error_dict:
            logging.getLogger("HWR").error(
                "%s: MacroServer not running?, %s", str(self.name()), error_dict
            )
            self.emit("commandFailed", (-1, self.name()))
        except Exception:
            logging.getLogger("HWR").exception(
                "%s: an error occured when calling Tango command %s",
                str(self.name()),
                self.macro_format,
            )
            self.emit("commandFailed", (-1, self.name()))

        if wait:
            logging.getLogger("HWR").debug("... start waiting...")
            t = gevent.spawn(end_of_macro, self)
            t.get()
            logging.getLogger("HWR").debug("... end waiting...")

        return

    def update(self, event):
        """update the macro command status: ``commandCanExecute``, ``commandReady``, ``commandNotReady``, ``commandReplyArrive``, ``commandReplyAbort`` and ``commandFailed``"""

        data = event.event[2]

        try:
            if not isinstance(data, PyTango.DeviceAttribute):
                # Events different than a value changed on attribute.  Taurus sends an event with attribute info
                # logging.getLogger('HWR').debug("==========. Got an event, but it is not an attribute . it is %s" % type(data))
                # logging.getLogger('HWR').debug("doorstate event. type is %s" % str(type(data)))
                return

            # Handling macro state changed event
            doorstate = str(data.value)
            logging.getLogger("HWR").debug(
                "doorstate changed. it is %s" % str(doorstate)
            )

            if doorstate != self.doorstate:
                self.doorstate = doorstate

                # logging.getLogger('HWR').debug("self.doorstate is %s" % self.canExecute())
                self.emit("commandCanExecute", (self.can_execute(),))

                if doorstate in ["ON", "ALARM"]:
                    # logging.getLogger('HWR').debug("Macroserver ready for commands")
                    self.emit("commandReady", ())
                else:
                    # logging.getLogger('HWR').debug("Macroserver busy ")
                    self.emit("commandNotReady", ())

            if self.macrostate == SardanaMacro.STARTED and doorstate == "RUNNING":
                # logging.getLogger('HWR').debug("Macro server is running")
                self.macrostate = SardanaMacro.RUNNING
            elif self.macrostate == SardanaMacro.RUNNING and (
                doorstate in ["ON", "ALARM"]
            ):
                logging.getLogger("HWR").debug("Macro execution finished")
                self.macrostate = SardanaMacro.DONE
                self.result = self.door.result
                self.emit("commandReplyArrived", (self.result, str(self.name())))
                if doorstate == "ALARM":
                    self.emit("commandAborted", (str(self.name()),))
                self._reply_arrived_event.set()
            elif (
                self.macrostate == SardanaMacro.DONE
                or self.macrostate == SardanaMacro.INIT
            ):
                # already handled in the general case above
                pass
            else:
                logging.getLogger("HWR").debug("Macroserver state changed")
                self.emit("commandFailed", (-1, str(self.name())))
        except ConnectionFailed:
            logging.getLogger("HWR").debug("Cannot connect to door %s" % self.doorname)
            self.emit("commandFailed", (-1, str(self.name())))
        except Exception:
            import traceback

            logging.getLogger("HWR").debug(
                "SardanaMacro / event handling problem. Uggh. %s"
                % traceback.format_exc()
            )
            self.emit("commandFailed", (-1, str(self.name())))

    def abort(self):
        if self.door is not None:
            logging.getLogger("HWR").debug("SardanaMacro / aborting macro")
            self.door.abortMacro()
            # self.emit('commandReady', ())

    def is_connected(self):
        return self.door is not None

    def can_execute(self):
        return self.door is not None and (self.doorstate in ["ON", "ALARM"])


class SardanaCommand(CommandObject):
    """SardanaCommand"""

    def __init__(self, name, command, taurusname=None, username=None, **kwargs):
        CommandObject.__init__(self, name, username, **kwargs)

        self.command = command
        self.taurusname = taurusname
        self.device = None

    def init_device(self):
        try:
            self.device = Device(self.taurusname)
        except DevFailed as traceback:
            last_error = traceback[-1]
            logging.getLogger("HWR").error(
                "%s: %s", str(self.name()), last_error["desc"]
            )
            self.device = None
        else:
            try:
                self.device.ping()
            except ConnectionFailed:
                self.device = None
                raise ConnectionError

    def __call__(self, *args, **kwargs):
        self.emit("commandBeginWaitReply", (str(self.name()),))

        if self.device is None:
            self.init_device()

        try:
            cmdObject = getattr(self.device, self.command)
            ret = cmdObject(*args)
        except DevFailed as error_dict:
            logging.getLogger("HWR").error(
                "%s: Tango, %s", str(self.name()), error_dict
            )
        except Exception:
            logging.getLogger("HWR").exception(
                "%s: an error occured when calling Tango command %s",
                str(self.name()),
                self.command,
            )
        else:
            self.emit("commandReplyArrived", (ret, str(self.name())))
            return ret
        self.emit("commandFailed", (-1, self.name()))

    def abort(self):
        pass

    def is_connected(self):
        return self.device is not None


class SardanaChannel(ChannelObject, SardanaObject):
    """Creates a Sardana Channel"""

    def __init__(
        self, name, attribute_name, username=None, uribase=None, polling=None, **kwargs
    ):
        super(SardanaChannel, self).__init__(name, username, **kwargs)

        class ChannelInfo(object):
            def __init__(self):
                super(ChannelInfo, self).__init__()

        self.attribute_name = attribute_name
        self.model = os.path.join(uribase, attribute_name)
        self.attribute = None

        self.value = None
        self.polling = polling

        self.info = ChannelInfo()
        self.info.minval = None
        self.info.maxval = None

        self.init_device()

    def init_device(self):
        try:
            self.attribute = Attribute(self.model)
            #
            # DIRTY FIX to make compatible taurus listeners and existence of Tango channels/commands
            # as defined in Command/Tango.py
            #
            # if self.attribute.__class__ == taurus.core.tango.tangoattribute.TangoAttribute:
            #    dev = self.attribute.getParentObj()
            #    dp = dev.getHWObj()
            #    try:
            #        dp.subscribe_event = dp._subscribe_event
            #    except AttributeError:
            #        pass
            # logging.getLogger("HWR").debug("initialized")
        except DevFailed as traceback:
            self.imported = False
            return

        # read information
        try:
            if int(taurus.Release.version[0]) == 3:
                ranges = self.attribute.getConfig().getRanges()
                if ranges is not None and ranges[0] != "Not specified":
                    self.info.minval = float(ranges[0])
                if ranges is not None and ranges[-1] != "Not specified":
                    self.info.maxval = float(ranges[-1])
            elif int(taurus.Release.version[0]) > 3:  # taurus 4 and beyond
                minval, maxval = self.attribute.getRange()
                if minval is not None:
                    self.info.minval = minval.magnitude
                if maxval is not None:
                    self.info.maxval = maxval.magnitude

        except Exception:
            import traceback

            logging.getLogger("HWR").info("info initialized. Cannot get limits")
            logging.getLogger("HWR").info("%s" % traceback.format_exc())

        # prepare polling
        # if the polling value is a number set it as the taurus polling period

        if self.polling:
            if isinstance(self.polling, int):
                self.attribute.changePollingPeriod(self.polling)

            self.attribute.addListener(self.object_listener)

    def get_value(self):
        return self._read_value()

    def set_value(self, new_value):
        self._write_value(new_value)

    def _write_value(self, new_value):
        self.attribute.write(new_value)

    def _read_value(self):
        value = self.attribute.read().value
        return value

    def get_info(self):
        try:
            limits = self.attribute.getLimits()
            self.info.minval = limits[0].magnitude
            self.info.maxval = limits[1].magnitude
        except Exception:
            import traceback

            logging.getLogger("HWR").info("%s" % traceback.format_exc())
        return self.info

    def update(self, event):
        data = event.event[2]

        try:
            new_value = data.value

            if new_value is None:
                new_value = self.get_value()

            if isinstance(new_value, tuple):
                new_value = list(new_value)

            self.value = new_value
            self.emit("update", self.value)
        except AttributeError:
            # No value in data... this is probably a connection error
            pass

    def is_connected(self):
        return self.attribute is not None

    def channel_listener(self, *args):
        ev = AttributeEvent(args)
        SardanaChannel._eventReceivers[id(ev)] = saferef.safe_ref(self.update)
        SardanaChannel._eventsQueue.put(ev)
        SardanaChannel._eventsProcessingTimer.send()



import logging

import gevent
import numpy
from PyQt4.QtGui import QApplication
from PyTango import DeviceProxy

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.Command.Tango import TangoCommand
from mxcubecore.HardwareObjects.abstract.AbstractMotor import MotorStates


class TangoDCMotor(HardwareObject):

    MOVESTARTED = 0
    NOTINITIALIZED = 0
    UNUSABLE = 0
    READY = 2
    MOVING = 4
    ONLIMIT = 1

    stateDict = {
        "UNKNOWN": 0,
        "OFF": 0,
        "ALARM": 1,
        "FAULT": 1,
        "STANDBY": 2,
        "RUNNING": 4,
        "MOVING": 4,
        "ON": 2,
    }

    def __init__(self, name):

        # State values as expected by Motor bricks

        super().__init__(name)
        self.GUIstep = 0.1
        self.motor_states = MotorStates()

    def _init(self):
        self.positionValue = 0.0
        self.stateValue = "UNKNOWN"

        threshold = self.get_property("threshold")
        self.threshold = (
            0.0018  # default value. change it with property threshold in xml
        )

        self.old_value = 0.0
        self.tangoname = self.get_property("tangoname")
        self.motor_name = self.get_property("motor_name")
        self.ho = DeviceProxy(self.tangoname)

        try:
            self.dataType = self.get_property("datatype")
            if self.dataType is None:
                self.dataType = "float"
        except Exception:
            self.dataType = "float"

        if threshold is not None:
            try:
                self.threshold = float(threshold)
            except Exception:
                pass

        self.set_is_ready(True)
        try:
            self.limitsCommand = self.get_command_object("limits")
        except KeyError:
            self.limitsCommand = None
        self.positionChan = self.get_channel_object(
            "position"
        )  # utile seulement si positionchan n'est pas defini dans le code
        self.stateChan = self.get_channel_object(
            "state"
        )  # utile seulement si statechan n'est pas defini dans le code

        self.positionChan.connect_signal("update", self.positionChanged)
        self.stateChan.connect_signal("update", self.motorStateChanged)

    def positionChanged(self, value):
        self.positionValue = value
        if abs(float(value) - self.old_value) > self.threshold:
            try:
                # logging.getLogger("HWR").error("%s: TangoDCMotor new position  , %s", self.name(), value)
                self.emit("valueChanged", (value,))
                self.old_value = value
            except Exception:
                logging.getLogger("HWR").error(
                    "%s: TangoDCMotor not responding, %s", self.id, ""
                )
                self.old_value = value

    def is_ready(self):
        return self.stateValue == "STANDBY"

    def connect_notify(self, signal):
        if signal == "hardware_object_name,stateChanged":
            self.motorStateChanged(TangoDCMotor.stateDict[self.stateValue])
        elif signal == "limitsChanged":
            self.motorLimitsChanged()
        elif signal == "valueChanged":
            self.motor_positions_changed(self.positionValue)
        self.set_is_ready(True)

    def motorState(self):
        return TangoDCMotor.stateDict[self.stateValue]

    def motorStateChanged(self, state):
        self.stateValue = str(state)
        self.set_is_ready(True)
        logging.info("motor state changed. it is %s " % self.stateValue)
        self.emit("stateChanged", (TangoDCMotor.stateDict[self.stateValue],))

    def get_state(self):
        return TangoDCMotor.stateDict[self.stateValue]

    def get_limits(self):
        try:
            logging.getLogger("HWR").info(
                "TangoDCMotor.get_limits: trying to get limits for motor_name %s "
                % (self.motor_name)
            )
            limits = self.ho.getMotorLimits(
                self.motor_name
            )  # limitsCommand() # self.ho.getMotorLimits(self.motor_name)
            logging.getLogger("HWR").info(
                "TangoDCMotor.get_limits: Getting limits for %s -- %s "
                % (self.motor_name, str(limits))
            )
            if numpy.inf in limits:
                limits = numpy.array([-10000, 10000])
        except Exception:
            # import traceback
            # logging.getLogger("HWR").info("TangoDCMotor.get_limits: Cannot get limits for %s.\nException %s " % (self.motor_name, traceback.print_exc()))
            if self.motor_name in [
                "detector_distance",
                "detector_horizontal",
                "detector_vertical",
            ]:
                info = self.positionChan.getInfo()
                limits = [float(info.min_value), float(info.max_value)]
            # if self.motor_name == 'detector_ts':
            # limits = [96, 1100]
            # elif self.motor_name == 'detector_tx':
            # limits =
            elif self.motor_name == "exposure":
                limits = [float(self.min_value), float(self.max_value)]

        if limits is None:
            try:
                limits = self.get_property("min"), self.get_property("max")
                logging.getLogger("HWR").info(
                    "TangoDCMotor.get_limits: %.4f ***** %.4f" % limits
                )
                limits = numpy.array(limits)
            except Exception:
                # logging.getLogger("HWR").info("TangoDCMotor.get_limits: Cannot get limits for %s" % self.name())
                limits = None
        return limits

    def motorLimitsChanged(self):
        self.emit("limitsChanged", (self.get_limits(),))

    def motorIsMoving(self):
        return self.stateValue == "RUNNING" or self.stateValue == "MOVING"

    def motorMoveDone(self, channelValue):
        if self.stateValue == "STANDBY":
            self.emit("moveDone", (self.tangoname, "tango"))

    def motor_positions_changed(self, absolutePosition):
        self.emit("valueChanged", (absolutePosition,))

    def syncQuestionAnswer(self, specSteps, controllerSteps):
        return (
            "0"  # This is only for spec motors. 0 means do not change anything on sync
        )

    def get_value(self):
        return self.positionChan.get_value()

    def convertValue(self, value):
        logging.info("TangoDCMotor: converting value to %s " % str(self.dataType))
        retvalue = value
        if self.dataType in ["short", "int", "long"]:
            retvalue = int(value)
        return retvalue

    def get_motor_mnemonic(self):
        return self.name

    def _set_value(self, value):
        """Move the motor to the required position

        Arguments:
        absolutePosition -- position to move to
        """
        logging.getLogger("TangoClient").info(
            "TangoDCMotor move (%s). Trying to go to %s: type '%s'",
            self.motor_name,
            value,
            type(value),
        )
        value = float(value)
        if not isinstance(value, float) and not isinstance(value, int):
            logging.getLogger("TangoClient").error(
                "Cannot move %s: position '%s' is not a number. It is a %s",
                self.tangoname,
                value,
                type(value),
            )
        logging.info("TangoDCMotor: move. motor will go to %s " % str(value))
        logging.getLogger("HWR").info(
            "TangoDCMotor.move to absolute position: %.3f" % value
        )
        logging.getLogger("TangoClient").info(
            "TangoDCMotor move. Trying to go to %s: that is a '%s'", value, type(value),
        )
        # if abs(self.get_value() - value) > epsilon:
        #     logging.info(
        #         "TangoDCMotor: difference larger then epsilon (%s), executing the move "
        #         % str(epsilon)
        #     )
        self.positionChan.set_value(self.convertValue(value))
        # else:
        #     logging.info(
        #         "TangoDCMotor: not moving really as epsilon is large %s " % str(epsilon)
        #     )
        #     logging.info("TangoDCMotor: self.get_value() %s " % str(self.get_value()))
        #     logging.info("TangoDCMotor: value %s " % str(value))

    def stop(self):
        logging.getLogger("HWR").info("TangoDCMotor.stop")
        stopcmd = self.get_command_object("Stop")()
        if not stopcmd:
            stopcmd = TangoCommand("stopcmd", "Stop", self.tangoname)
        stopcmd()

    def isSpecConnected(self):
        logging.getLogger().debug("%s: TangoDCMotor.isSpecConnected()" % self.name())
        return (Truehardware_object_name,)


def test():
    hwr = HWR.get_hardware_repository()
    hwr.connect()

    motor = hwr.get_hardware_object("/phi")
    print(motor.get_value())


if __name__ == "__main__":
    test()



import ast
import importlib
import logging
import operator
import sys

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.CommandContainer import (
    ARGUMENT_TYPE_LIST,
    TWO_STATE_COMMAND_T,
    CommandObject,
)
from mxcubecore.TaskUtils import task
from mxcubecore.utils.conversion import camel_to_snake


class ControllerCommand(CommandObject):
    def __init__(self, name, cmd=None, username=None, klass=None):
        CommandObject.__init__(self, name, username)

        if not cmd:
            self._cmd = klass()
        else:
            self._cmd = cmd

        self._cmd_execution = None
        self.type = "CONTROLLER"

        if self.name() == "Anneal":
            self.add_argument("Time [s]", "float")
        if self.name() == "Test":
            self.add_argument("combo test", "combo", [{"value1": 0, "value2": 1}])

    def is_connected(self):
        return True

    @task
    def __call__(self, *args, **kwargs):
        """Call the command"""
        self.emit("commandBeginWaitReply", (str(self.name()),))
        self._cmd_execution = gevent.spawn(self._cmd, *args, **kwargs)
        self._cmd_execution.link(self._cmd_done)

    def _cmd_done(self, cmd_execution):
        """Handle the command execution.
        Args:
            (obj): Command execution greenlet.
        """
        try:
            try:
                res = cmd_execution.get()
                res = res if res else ""
            except Exception:
                logging.getLogger("HWR").exception(
                    "%s: execution failed", str(self.name())
                )
                self.emit("commandFailed", (str(self.name()),))
            else:
                if isinstance(res, gevent.GreenletExit):
                    # command aborted
                    self.emit("commandFailed", (str(self.name()),))
                else:
                    self.emit("commandReplyArrived", (str(self.name()), res))
        finally:
            self.emit("commandReady", (str(self.name()), ""))

    def abort(self):
        """Abort the execution."""
        if self._cmd_execution and not self._cmd_execution.ready():
            self._cmd_execution.kill()

    def value(self):
        return None


class HWObjActuatorCommand(CommandObject):
    """Class for two state hardware objects"""

    def __init__(self, name, hwobj):
        super().__init__(name)
        self._hwobj = hwobj
        self.type = TWO_STATE_COMMAND_T
        self.argument_type = ARGUMENT_TYPE_LIST
        self._hwobj.connect("valueChanged", self._cmd_done)
        self._running = False

    def _get_action(self):
        """Return which action has to be executed.
        Return:
            (str): The name of the command
        """
        values = [v.name for v in self._hwobj.VALUES]
        values.remove("UNKNOWN")
        values.remove(self._hwobj.get_value().name)

        return self._hwobj.VALUES[values[0]]

    @task
    def __call__(self, *args, **kwargs):
        """Execute the action.
        Args: None
        Kwargs: None
        """
        self._running = True
        self.emit("commandBeginWaitReply", (str(self.name()),))
        value = self._get_action()
        self._hwobj.set_value(value, timeout=60)

    def _cmd_done(self, state):
        """Handle the command execution.
        Args:
            (obj): Command execution greenlet.
        """
        gevent.sleep(1)
        try:
            res = self._hwobj.get_value().name
        except Exception:
            self.emit("commandFailed", (str(self.name()),))
        else:
            if isinstance(res, gevent.GreenletExit):
                self.emit("commandFailed", (str(self.name()),))
            elif self._running:
                self.emit("commandReplyArrived", (str(self.name()), res))

        self._running = False

    def value(self):
        """Return the current command vaue.
        Return:
            (str): The value as a string
        """
        value = "UNKNOWN"
        if hasattr(self._hwobj, "get_value"):
            value = self._hwobj.get_value()
            try:
                return value.name
            except AttributeError:
                return value
        return value


class AnnotatedCommand(CommandObject):
    def __init__(self, beamline_action_ho, name, cmd_name):
        self._beamline_action_ho = beamline_action_ho
        self._name = name
        self._cmd_name = cmd_name
        self._value = ""
        self._last_result = None
        self._messages = []
        self.task = None

    def get_value(self):
        return self._value

    def set_last_result(self, result):
        self._last_result = result

    @property
    def cmd_name(self):
        return self._cmd_name


class BeamlineActions(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)
        self._annotated_commands = []
        self._annotated_command_dict = {}
        self._command_list = []
        self._current_command = None

    def _get_command_object_class(self, path_str):
        parts = path_str.split(".")

        _module_name = "mxcubecore." + ".".join(parts[:-1])
        _cls_name = parts[-1]
        self._annotated_commands.append(_cls_name)

        # assume import from current module if only class name given (no module)
        if len(parts) == 1:
            _cls = getattr(sys.modules[__name__], _cls_name)
        else:
            _mod = importlib.import_module(_module_name)
            _cls = getattr(_mod, _cls_name)

        return _cls

    def init(self):
        command_list = ast.literal_eval(
            self.get_property("commands").strip().replace("\n", "")
        )

        for command in command_list:
            attrname = camel_to_snake(command["command"].split(".")[-1])

            if hasattr(self, attrname):
                msg = (
                    "Command with name %s already exists"
                    % command["command"].split(".")[-1]
                )
                logging.getLogger("HWR").warning(msg)
                continue

            if command["type"] == "annotated":
                _cls = self._get_command_object_class(command["command"])
                fname = camel_to_snake(_cls.__name__)
                _cls_inst = _cls(self, fname.replace("_", " ").title(), fname)

                self._annotated_command_dict[fname] = _cls_inst
                setattr(self, attrname, getattr(_cls_inst, fname))
                self._exports_config_list.append(attrname)
            elif command["type"] == "controller":
                try:
                    cmd = operator.attrgetter(command["command"])(self)
                    _cmd_obj = ControllerCommand(command["name"], cmd, command["name"])
                except AttributeError:
                    _cls = self._get_command_object_class(command["command"])
                    _cmd_obj = ControllerCommand(
                        command["name"], None, command["name"], klass=_cls
                    )

                self._command_list.append(_cmd_obj)
                setattr(self, attrname, _cmd_obj)
            elif command["type"] == "actuator":
                try:
                    cmd = operator.attrgetter(command["command"])
                    _cmd_obj = HWObjActuatorCommand(command["name"], cmd)
                    self._command_list.append(_cmd_obj)
                except AttributeError:
                    pass

        super().init()

    def get_annotated_command(self, name):
        return self._annotated_command_dict[name]

    def get_annotated_commands(self):
        return list(self._annotated_command_dict.values())

    def _execute_annotated_command(self, name, args):
        cmd = getattr(self, name, None)
        self._annotated_command_dict[name].emit("commandBeginWaitReply", name)

        _model = self.pydantic_model[name](**args)
        _all_child_models = []

        for _key in _model.dict().keys():
            _all_child_models.append(getattr(_model, _key))

        _t = gevent.spawn(cmd, *_all_child_models)
        _t.link(self._command_done)

        self._current_command = cmd.__self__
        self._current_command.task = _t

    def get_commands(self):
        return self._command_list

    def _execute_command(self, name, args):
        try:
            cmds = self.get_commands()
        except Exception:
            cmds = []

        for cmd in cmds:
            if cmd.name() == name:
                try:
                    cmd.emit("commandBeginWaitReply", name)
                    logging.getLogger("user_level_log").info(
                        "Starting %s(%s)",
                        cmd.name(),
                        ", ".join(map(str, args)),
                    )
                    cmd(*args)
                except Exception as ex:
                    err = str(sys.exc_info()[1])
                    raise Exception(str(err)) from ex

    def execute_command(self, name, args):
        cmd = getattr(self, name, None)

        if cmd:
            self._execute_annotated_command(name, args)
        else:
            self._execute_command(name, args)

    def _command_done(self, greenlet):
        cmd_obj = self._annotated_command_dict[self._current_command.cmd_name]
        result = ""

        try:
            result = greenlet.get()
        except Exception:
            logging.getLogger("HWR").exception(
                "%s: execution failed", self._current_command.cmd_name
            )
            cmd_obj.emit("commandFailed", (self._current_command.cmd_name,))
        else:
            self._current_command.set_last_result(result)
            if isinstance(result, gevent.GreenletExit):
                # command aborted
                cmd_obj.emit("commandFailed", (self._current_command.cmd_name,))
                result = ""
            else:
                cmd_obj.emit(
                    "commandReplyArrived", (self._current_command.cmd_name, result)
                )
        finally:
            cmd_obj.emit(
                "commandReady",
                (self._current_command.cmd_name, result),
            )

            self._current_command = None

    def abort_command(self, name):
        if (
            self._current_command
            and self._current_command.cmd_name in self._annotated_command_dict
        ):
            self._annotated_command_dict[self._current_command.cmd_name]
            self._current_command.task.kill()
        else:
            for cmd in self.get_commands():
                if cmd.name() == name:
                    cmd.abort()
                    break



#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


from enum import (
    Enum,
    unique,
)

import gevent

from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractShutter

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


@unique
class ShutterValueEnum(Enum):
    """Defines only the compulsory values."""

    OPEN = "Open"
    CLOSED = "Closed"
    UNKNOWN = "Unknown"
    DISABLED = "Disabled"


class MDFastShutter(AbstractShutter):
    """
    MD Fast shutter
    """

    VALUES = ShutterValueEnum

    def __init__(self, name):
        """
        init
        :param name:
        """
        super(MDFastShutter, self).__init__(name)

        self._nominal_limits = None

        self.chan_shutter_is_open = None
        self.chan_current_phase = None

        self.shutter_is_open = None
        self.current_phase = None

    def init(self):
        super(MDFastShutter, self).init()

        self.chan_shutter_is_open = self.get_channel_object("chanShutterIsOpen")
        if self.chan_shutter_is_open:
            self.chan_shutter_is_open.connect_signal("update", self.shutter_is_open_changed)

        self.chan_current_phase = self.get_channel_object("chanCurrentPhase")
        if self.chan_current_phase is not None:
            self.current_phase = self.chan_current_phase.get_value()
            self.connect(self.chan_current_phase, "update", self.current_phase_changed)

    def shutter_is_open_changed(self, value):
        """
        Shutter state changed event
        :param value:
        :return:
        """
        self.shutter_is_open = value
        #We allow to control the fast shutter just in the beam location phase
        if self.current_phase == "BeamLocation":
            if value:
                value = self.VALUES.OPEN
            else:
                value = self.VALUES.CLOSED
        else:
            value = self.VALUES.DISABLED
        print("is open changed ", value, self.shutter_is_open)
        self.update_value(value)

    def get_value(self):
        return self._nominal_value

    def current_phase_changed(self, value):
        """
        Phase changed
        :param value: str
        :return:
        """
        self.current_phase = value
        self.shutter_is_open_changed(self.chan_shutter_is_open.get_value())

    def _set_value(self, value):
        if value == self.VALUES.OPEN:
            self.chan_shutter_is_open.set_value(True)
        elif value == self.VALUES.CLOSED:
            self.chan_shutter_is_open.set_value(False)

    def open_delete(self, wait=True):
        """
        Opens the shutter
        :param wait:
        :return:
        """
        self.chan_shutter_is_open.set_value(True)
        with gevent.Timeout(10, Exception("Timeout waiting for fast shutter open")):
            while not self.shutter_is_open:
                gevent.sleep(0.1)

    def close_delete(self, wait=True):
        """
        Closes shutter
        :param wait: boolean
        :return:
        """
        self.chan_shutter_is_open.set_value(False)
        with gevent.Timeout(10, Exception("Timeout waiting for fast shutter close")):
            while self.shutter_is_open:
                gevent.sleep(0.1)



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""
ESRF XRF scan procedure
Example xml_ configuration:

.. code-block:: xml

 <object class="ESRF.ESRFXRFSpectrum">
   <object href="/bliss" role="controller"/>
   <cfgfile>/users/blissadm/local/beamline_configuration/misc/15keV.cfg</cfgfile
>
   <default_integration_time>3</default_integration_time>
   <default_energy_range>[2.0, 15]</default_energy_range>
   <cfg_energies>[15]</cfg_energies>
 </object>
"""

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"

import logging
import os.path
from ast import literal_eval
from unittest.mock import MagicMock
from warnings import warn

import numpy
from PyMca5.PyMca import (
    ClassMcaTheory,
    ConfigDict,
    QtMcaAdvancedFitReport,
)

from mxcubecore.HardwareObjects.abstract.AbstractXRFSpectrum import AbstractXRFSpectrum

# Next line is a trick to avoid core dump in QtMcaAdvancedFitReport
QtMcaAdvancedFitReport.qt = MagicMock()


class ESRFXRFSpectrum(AbstractXRFSpectrum):
    """ESRF implementation of the XRF spectrum procedure"""

    def __init__(self, name):
        super().__init__(name)
        self.cfgfile = None
        self.config_fit = None
        self.ctrl_hwobj = None
        self.mcafit = None
        self.default_erange = None
        self.cfg_energies = []

    def init(self):
        """Initialise the objects and the properties"""
        super().init()
        self.ctrl_hwobj = self.get_object_by_role("controller")
        self.cfgfile = self.get_property(
            "cfgfile", "/users/blissadm/local/beamline_configuration/misc/15keV.cfg"
        )
        self.config_fit = ConfigDict.ConfigDict()
        self.mcafit = ClassMcaTheory.McaTheory(self.cfgfile)
        self.default_erange = literal_eval(
            self.get_property("default_energy_range", "[2.0, 15.0]")
        )
        self.cfg_energies = literal_eval(
            self.get_property("cfg_energies", "[7, 9, 12, 15]")
        )

    def _doSpectrum(self, ctime, filename):
        """Deprecated method."""
        warn(
            "_doSpectrum is deprecated. use _execute_spectrum",
            DeprecationWarning,
        )
        return self._execute_spectrum(ctime, filename)

    def _execute_spectrum(self, integration_time=None, filename=None):
        """Local XRF spectrum sequence.

        Args:
            integration_time (float): MCA integration time [s].
            filename (str): Data file (full path).
        Returns:
            (bool): Procedure executed correcly (True) or error (False)
        """
        filename = filename or self.spectrum_info_dict["filename"]
        integration_time = integration_time or self.default_integration_time

        # protect the detector
        self.ctrl_hwobj.detcover.set_in(20)
        try:
            current_transm = self.ctrl_hwobj.find_max_attenuation(
                ctime=float(integration_time),
                datafile=filename,
                roi=self.default_erange,
            )
            self.spectrum_info_dict["beamTransmission"] = current_transm
        except Exception as exp:
            logging.getLogger("user_level_log").exception(str(exp))
            self.spectrum_command_failed()
            return False

        # put away the fluo detector
        self.ctrl_hwobj.diffractometer.fldet_out()
        return True

    # Next methods are for fitting the data with pymca

    def mcafit_configuration(self, config=None):
        """Configure the fitting parameters. The procedure is time consuming.
           It is only executed if the last configuration file is not the same.

        Args:
            config(dict): Configuration dictionary, containing among others the
                          configuration file name.
        """
        change = False
        if not config or "file" not in config:
            cfgfile = self._get_cfgfile(self.spectrum_info_dict["energy"])
        else:
            cfgfile = config["file"]

        if self.cfgfile != cfgfile:
            self.cfgfile = cfgfile
            change = True
        self.config_fit.read(self.cfgfile)
        if "concentrations" not in self.config_fit:
            self.config_fit["concentrations"] = {}
            change = True
        if "attenuators" not in self.config_fit:
            self.config_fit["attenuators"] = {
                "Matrix": [1, "Water", 1.0, 0.01, 45.0, 45.0]
            }
            change = True
        if "flux" in config:
            self.config_fit["concentrations"]["flux"] = float(config["flux"])
            change = True
        if "time" in config:
            self.config_fit["concentrations"]["time"] = float(config["time"])
            change = True

        if change:
            self.mcafit.configure(self.config_fit)

    def spectrum_analyse(self, data=None, calib=None, config=None):
        """Execute the fitting. Write the fitted data files to the archive
        directory.

        Args:
            data (list): The raw data.
            calib (list): The mca calibration.
            config (dict): The configuration dictionary.
        """

        if not config:
            config = {
                "energy": self.spectrum_info_dict["energy"],
                "att": self.spectrum_info_dict["beamTransmission"],
                "bsX": self.spectrum_info_dict["beamSizeHorizontal"],
                "bsY": self.spectrum_info_dict["beamSizeVertical"],
                "legend": self.spectrum_info_dict["annotatedPymcaXfeSpectrum"],
                "htmldir": os.path.split(
                    self.spectrum_info_dict["annotatedPymcaXfeSpectrum"]
                )[0],
            }

        self.mcafit_configuration(config)
        calib = calib or self.ctrl_hwobj.mca.calibration

        # the spectrum is read by the find_max_attenuation procedure.
        # We only need the data, do not to read it again
        data = data or self.ctrl_hwobj.mca.data

        try:
            if data[0].size == 2:
                xdata = numpy.array(data[:, 0]) * 1.0
                ydata = numpy.array(data[:, 1])
            else:
                xdata = data[0] * 1.0
                ydata = data[1]

            try:
                xmin = self.config_fit["fit"]["xmin"]
                xmax = self.config_fit["fit"]["xmax"]
            except KeyError:
                xmin = data[0][0]
                xmax = data[0][-1]
            self.mcafit.setData(xdata, ydata, xmin=xmin, xmax=xmax, calibration=calib)

            self.mcafit.estimate()

            fitresult = self.mcafit.startfit(digest=1)
            if fitresult:
                fitresult = {"fitresult": fitresult[0], "result": fitresult[1]}

                # write the csv file to pyarch
                csvname = self.spectrum_info_dict["fittedDataFileFullPath"]
                self._write_csv_file(fitresult["result"], csvname)

                # write html report to pyarch
                fname = os.path.basename(self.spectrum_info_dict["filename"])
                outfile = fname.split(".")[0]
                outdir = os.path.dirname(
                    self.spectrum_info_dict["annotatedPymcaXfeSpectrum"]
                )

                _kw = {
                    "outdir": outdir,
                    "outfile": outfile,
                    "fitresult": fitresult,
                    "plotdict": {"logy": False},
                }

                report = QtMcaAdvancedFitReport.QtMcaAdvancedFitReport(**_kw)
                text = report.getText()
                report.writeReport(text=text)
                return True
            return False
        except Exception as exp:
            msg = f"XRFSpectrum: problem fitting {exp}\nPlease check the raw "
            msg += f"data file {self.spectrum_info_dict['scanFileFullPath']}"
            logging.getLogger("user_level_log").exception(msg)
            self.spectrum_command_failed()
            return False

    def _write_csv_file(self, fitresult, fname=None):
        """Write fitted data to a csv file.

        Args:
            fitresult(dict): Data as dictionary.
        Kwargs:
            fname (str): Filename to write to (full path).
        """
        fname = fname or self.spectrum_info_dict["fittedDataFileFullPath"]
        if os.path.exists(fname):
            os.remove(fname)

        # get the significant peaks
        peaks_dict = {}
        pars_len = len(fitresult["parameters"])
        grp_len = len(fitresult["groups"])
        nglobal = pars_len - grp_len
        parameters = fitresult["fittedpar"][:nglobal] + [0.0] * grp_len

        for grp in fitresult["parameters"][nglobal:]:
            idx = fitresult["parameters"].index(grp)
            parameters[idx] = fitresult["fittedpar"][idx]
            xmatrix = fitresult["xdata"]
            ymatrix = self.mcafit.mcatheory(parameters, xmatrix)
            ymatrix.shape = [len(ymatrix), 1]
            label = "y" + grp
            if self.mcafit.STRIP:
                peaks_dict[label] = ymatrix + self.mcafit.zz
            else:
                peaks_dict[label] = ymatrix
            peaks_dict[label].shape = (len(peaks_dict[label]),)
            parameters[idx] = 0.0

        delimiter = ","
        header = f'"channel"{delimiter}"Energy"{delimiter}"counts"{delimiter}'
        header += f'"fit"{delimiter}"continuum"{delimiter}"pileup"'

        # add the peaks label
        for key in peaks_dict:
            header += delimiter + f'"{key}"'

        # logging.getLogger("user_level_log").info("Writing %s" % fname)
        with open(fname, "w") as csv_fd:
            csv_fd.write(header)
            csv_fd.write("\n")
            for i in range(fitresult["xdata"].size):
                csv_fd.write(
                    "%.7g%s%.7g%s%.7g%s%.7g%s%.7g%s%.7g"
                    % (
                        fitresult["xdata"][i],
                        delimiter,
                        fitresult["energy"][i],
                        delimiter,
                        fitresult["ydata"][i],
                        delimiter,
                        fitresult["yfit"][i],
                        delimiter,
                        fitresult["continuum"][i],
                        delimiter,
                        fitresult["pileup"][i],
                    )
                )
                for val in peaks_dict.values():
                    csv_fd.write(f"{delimiter}{val[i]:.7g}")

                csv_fd.write("\n")

    def _get_cfgfile(self, energy):
        """Get the correct configuration file.

        Args:
            energy(float): The energy to choose which configuration file.
        """
        self.cfg_energies.sort()

        cfg_path = os.path.split(self.cfgfile)[0]
        for egy in self.cfg_energies:
            if egy > energy:
                return os.path.join(cfg_path, f"{str(egy)}keV.cfg")
        return self.cfgfile



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent
import gevent.event

try:
    import Queue as queue
except ImportError:
    import queue

import numpy

from mxcubecore import Poller
from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
    ConnectionError,
)
from mxcubecore.dispatcher import saferef

gevent_version = list(map(int, gevent.__version__.split(".")))

log = logging.getLogger("HWR")

try:
    import PyTango
    from PyTango.gevent import DeviceProxy
except ImportError:
    logging.getLogger("HWR").warning("Tango support is not available.")


__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"

log = logging.getLogger("HWR")


class TangoCommand(CommandObject):
    def __init__(self, name, command, tangoname=None, username=None, **kwargs):
        CommandObject.__init__(self, name, username, **kwargs)

        self.command = command
        self.device_name = tangoname
        self.device = None

    def init_device(self):
        try:
            self.device = DeviceProxy(self.device_name)
        except PyTango.DevFailed as traceback:
            last_error = traceback[-1]
            logging.getLogger("HWR").error(
                "%s: %s", str(self.name()), last_error["desc"]
            )
            self.device = None
        else:
            try:
                self.device.ping()
            except PyTango.ConnectionFailed:
                self.device = None
                raise ConnectionError

    def __call__(self, *args, **kwargs):
        self.emit("commandBeginWaitReply", (str(self.name()),))
        if self.device is None:
            # TODO: emit commandFailed
            # beware of infinite recursion with Sample Changer
            # (because of procedure exception cleanup...)
            self.init_device()

        try:
            tango_cmd_object = getattr(self.device, self.command)
            ret = tango_cmd_object(
                *args
            )  # eval('self.device.%s(*%s)' % (self.command, args))
        except PyTango.DevFailed as error_dict:
            logging.getLogger("HWR").error(
                "%s: Tango, %s", str(self.name()), error_dict
            )
        except Exception:
            logging.getLogger("HWR").exception(
                "%s: an error occured when calling Tango command %s",
                str(self.name()),
                self.command,
            )
        else:
            self.emit("commandReplyArrived", (ret, str(self.name())))
            return ret
        self.emit("commandFailed", (-1, self.name()))

    def abort(self):
        pass

    def set_device_timeout(self, timeout):
        if self.device is None:
            self.init_device()
        self.device.set_timeout_millis(timeout)

    def is_connected(self):
        return self.device is not None


def process_tango_events():
    while not TangoChannel._tangoEventsQueue.empty():
        try:
            ev = TangoChannel._tangoEventsQueue.get_nowait()
        except queue.Empty:
            break
        else:
            try:
                receiverCbRef = TangoChannel._eventReceivers[id(ev)]
                receiverCb = receiverCbRef()
                if receiverCb is not None:
                    try:
                        gevent.spawn(receiverCb, ev.event.attr_value.value)
                    except AttributeError:
                        pass
            except KeyError:
                pass


class E:
    def __init__(self, event):
        self.event = event


class TangoChannel(ChannelObject):
    _tangoEventsQueue = queue.Queue()
    _eventReceivers = {}

    _tangoEventsProcessingTimer = gevent.get_hub().loop.async_()

    # start Tango events processing timer
    _tangoEventsProcessingTimer.start(process_tango_events)

    def __init__(
        self,
        name,
        attribute_name,
        tangoname=None,
        username=None,
        polling=None,
        timeout=10000,
        **kwargs,
    ):
        ChannelObject.__init__(self, name, username, **kwargs)

        self.attribute_name = attribute_name
        self.device_name = tangoname
        self.device = None
        self.value = Poller.NotInitializedValue
        self.polling = polling
        self.polling_timer = None
        self.polling_events = False
        self.timeout = int(timeout)
        self.read_as_str = kwargs.get("read_as_str", False)
        self._device_initialized = gevent.event.Event()
        self.init_device()
        self.continue_init(None)

    def init_poll_failed(self, e, poller_id):
        self._device_initialized.clear()
        logging.warning(
            "%s/%s (%s): could not complete init. (hint: device server is not running, or has to be restarted)",
            self.device_name,
            self.attribute_name,
            self.name(),
        )
        self.init_poller = self.init_poller.restart(3000)

    def continue_init(self, _):
        if isinstance(self.polling, int):
            self.raw_device = DeviceProxy(self.device_name)

            Poller.poll(
                self.poll,
                polling_period=self.polling,
                value_changed_callback=self.update,
                error_callback=self.poll_failed,
            )
        else:
            if self.polling == "events":
                # try to register event
                try:
                    self.polling_events = True
                    self.device.subscribe_event(
                        self.attribute_name,
                        PyTango.EventType.CHANGE_EVENT,
                        self,
                        [],
                        True,
                    )
                except Exception:
                    logging.getLogger("HWR").exception("could not subscribe event")
        self._device_initialized.set()

    def init_device(self):
        try:
            self.device = DeviceProxy(self.device_name)
        except PyTango.DevFailed as traceback:
            self.imported = False
            last_error = traceback[-1]
            logging.getLogger("HWR").error(
                "%s: %s", str(self.name()), last_error["desc"]
            )
        else:
            self.imported = True
            try:
                self.device.ping()
            except PyTango.ConnectionFailed:
                self.device = None
                raise ConnectionError
            else:
                self.device.set_timeout_millis(self.timeout)

                # check that the attribute exists (to avoid Abort in PyTango grrr)
                if not self.attribute_name.lower() in [
                    attr.name.lower() for attr in self.device.attribute_list_query()
                ]:
                    logging.getLogger("HWR").error(
                        "no attribute %s in Tango device %s",
                        self.attribute_name,
                        self.device_name,
                    )
                    self.device = None

    def push_event(self, event):
        if (
            event.attr_value is None
            or event.err
            or event.attr_value.quality != PyTango.AttrQuality.ATTR_VALID
        ):
            return
        else:
            pass

        ev = E(event)
        TangoChannel._eventReceivers[id(ev)] = saferef.safe_ref(self.update)
        TangoChannel._tangoEventsQueue.put(ev)
        TangoChannel._tangoEventsProcessingTimer.send()

    def poll(self):
        def read_attr():
            if self.read_as_str:
                value = self.raw_device.read_attribute(
                    self.attribute_name, PyTango.DeviceAttribute.ExtractAs.String
                ).value
            else:
                value = self.raw_device.read_attribute(self.attribute_name).value

            return value

        while True:
            try:  # in case of tango communication errors, retry reading the attribute
                return read_attr()
            except PyTango.DevFailed:
                log.warning(
                    f"error polling {self.raw_device} {self.attribute_name} attribute, retrying.",
                    exc_info=True,
                )
                gevent.sleep(0.1)
            except Exception:
                log.exception(
                    "unexpected exception polling %s %s attribute",
                    self.raw_device,
                    self.attribute_name,
                )
                raise

    def poll_failed(self, e, poller_id):
        self.emit("update", None)
        """
        emit_update = True
        if self.value is None:
          emit_update = False
        else:
          self.value = None

        try:
            self.init_device()
        except:
            pass

        poller = Poller.get_poller(poller_id)
        if poller is not None:
            poller.restart(1000)

        try:
          raise e
        except:
          logging.exception("%s: Exception happened while polling %s", self.name(), self.attribute_name)

        if emit_update:
          # emit at the end => can raise exceptions in callbacks
          self.emit('update', None)
        """

    def get_info(self):
        self._device_initialized.wait(timeout=3)
        return self.device.get_attribute_config(self.attribute_name)

    def update(self, value=Poller.NotInitializedValue):
        # start with checking if we have a numpy array, as comparing
        # numpy.ndarray to Poller.NotInitializedValue raises a ValueError exception
        if isinstance(value, numpy.ndarray):
            value = value.tolist()
        elif value == Poller.NotInitializedValue:
            value = self.get_value()
        elif isinstance(value, tuple):
            value = list(value)

        self.value = value
        self.emit("update", value)

    def get_value(self):
        if self.read_as_str:
            value = self.device.read_attribute(
                self.attribute_name, PyTango.DeviceAttribute.ExtractAs.String
            ).value
        else:
            value = self.device.read_attribute(self.attribute_name).value

        if isinstance(value, numpy.ndarray):
            if not numpy.array_equal(value, self.value):
                self.update(value)
        elif value != self.value:
            self.update(value)

        return value

    def set_value(self, new_value):
        self.device.write_attribute(self.attribute_name, new_value)

    def is_connected(self):
        return self.device is not None



#!/usr/bin/env python
# called this way
# autoPROCLauncher.py -path
# /data/staff/biomax/staff/jie/2015_11_10/processed -mode after
# -datacollectionID 13 -residues 200 -anomalous False -cell "0,0,0,0,0,0"

import logging
import os
import shlex
import string
import subprocess
import sys
import tempfile
import time
import urllib

import httplib

inputTemplate = """<?xml version="1.0"?>
<XSDataInputControlAutoPROC>
  <dataCollectionId>
    <value>{dataCollectionId}</value>
  </dataCollectionId>
  <doAnomAndNonanom>
    <value>{doAnomAndNonanom}</value>
  </doAnomAndNonanom>
  <processDirectory>
    <path>
      <value>{autoPROCPath}</value>
    </path>
  </processDirectory>
</XSDataInputControlAutoPROC>
"""

SCRIPT_TEMPLATE = """#!/usr/bin/env python

import os
import sys
import time
import socket
import traceback

sys.path.insert(0, "/mxn/groups/biomax/cmxsoft/edna-mx/kernel/src")

from EDVerbose import EDVerbose
from EDFactoryPluginStatic import EDFactoryPluginStatic

beamline = "$beamline"
proposal = "$proposal"
dataCollectionId = $dataCollectionId
autoPROCDirectory = "$autoPROCDirectory"
inputFile = "$inputFile"

pluginName = "EDPluginControlAutoPROCv1_0"
os.environ["EDNA_SITE"] = "MAXIV_BIOMAX"
os.environ["ISPyB_user"]=""
os.environ["ISPyB_pass"]=""

EDVerbose.screen("Executing EDNA plugin %s" % pluginName)
EDVerbose.screen("EDNA_SITE %s" % os.environ["EDNA_SITE"])

hostname = socket.gethostname()
dateString  = time.strftime("%Y%m%d", time.localtime(time.time()))
timeString = time.strftime("%H%M%S", time.localtime(time.time()))
#strPluginBaseDir = os.path.join("/tmp", beamline, dateString)
#if not os.path.exists(strPluginBaseDir):
#    os.makedirs(strPluginBaseDir, 0o755)

#baseName = "{0}_autoPROC".format(timeString)
#baseDir = os.path.join(strPluginBaseDir, baseName)
baseName = "{hostname}_{date}-{time}".format(hostname=hostname,
                                             date=dateString,
                                             time=timeString)
baseDir = os.path.join(autoPROCDirectory, baseName)
if not os.path.exists(baseDir):
    os.makedirs(baseDir, 0o755)
EDVerbose.screen("EDNA plugin working directory: %s" % baseDir)

#linkName = "{hostname}_{date}-{time}".format(hostname=hostname,
#                                             date=dateString,
#                                             time=timeString)
#os.symlink(baseDir, os.path.join(autoPROCDirectory, linkName))

ednaLogName = "autoPROC_{0}-{1}.log".format(dateString, timeString)
EDVerbose.setLogFileName(os.path.join(autoPROCDirectory, ednaLogName))
EDVerbose.setVerboseOn()

edPlugin = EDFactoryPluginStatic.loadPlugin(pluginName)
edPlugin.setDataInput(open(inputFile).read())
#edPlugin.setBaseDirectory(strPluginBaseDir)
edPlugin.setBaseDirectory(autoPROCDirectory)
edPlugin.setBaseName(baseName)

EDVerbose.screen("Start of execution of EDNA plugin %s" % pluginName)
os.chdir(baseDir)
edPlugin.executeSynchronous()

"""
# XDS.INP creation is now asynchronous in mxcube, so it may not be here yet
# when we're started
# WAIT_XDS_TIMEOUT = 10

HPC_HOST = "b-picard07-clu0-fe-0.maxiv.lu.se"


class AutoProcLauncher:
    def __init__(
        self,
        path,
        mode,
        datacollectionID,
        residues=200,
        anomalous=False,
        cell="0,0,0,0,0,0",
        spacegroup=None,
    ):
        self.autoprocessingPath = path
        self.mode = mode
        self.dataCollectionId = datacollectionID
        self.nres = residues
        self.anomalous = anomalous
        self.cell = cell
        self.spacegroup = spacegroup
        self.log = logging.getLogger()
        self.log.setLevel(logging.DEBUG)
        self.xdsAppeared = False
        self.autoPROCPath = os.path.join(self.autoprocessingPath, "autoPROC")
        if not os.path.exists(self.autoPROCPath):
            os.makedirs(self.autoPROCPath, 0o755)
        self.xdsInputFile = os.path.join(self.autoprocessingPath, "XDS.INP")
        self.doAnomAndNonanom = True

    def parse_input_file(self):
        # the other parameters are not used right now
        self.inputXml = inputTemplate.format(
            dataCollectionId=self.dataCollectionId,
            doAnomAndNonanom=self.doAnomAndNonanom,
            autoPROCPath=self.autoPROCPath,
        )

        # we now need a temp file in the data dir to write the data model to
        ednaInputFileName = "autoPROC_input.xml"
        self.ednaInputFilePath = os.path.join(self.autoPROCPath, ednaInputFileName)
        if os.path.exists(self.ednaInputFilePath):
            # Create unique file name
            ednaInputFile = tempfile.NamedTemporaryFile(
                suffix=".xml",
                prefix="autoPROC_input-",
                dir=self.autoPROCPath,
                delete=False,
            )
            self.ednaInputFilePath = os.path.join(self.autoPROCPath, ednaInputFile.name)
            ednaInputFile.file.write(self.inputXml)
            ednaInputFile.close()
        else:
            open(self.ednaInputFilePath, "w").write(self.inputXml)
        os.chmod(self.ednaInputFilePath, 0o755)

        directories = self.autoprocessingPath.split(os.path.sep)
        try:
            beamline = directories[3]
            proposal = directories[4]
        except Exception:
            beamline = "unknown"
            proposal = "unknown"

        # to do restrict autoPROC only for academic users!?

        template = string.Template(SCRIPT_TEMPLATE)
        self.script = template.substitute(
            beamline=beamline,
            proposal=proposal,
            autoPROCDirectory=self.autoPROCPath,
            dataCollectionId=self.dataCollectionId,
            inputFile=self.ednaInputFilePath,
        )

        # we also need some kind of script to run edna-plugin-launcher
        ednaScriptFileName = "autoPROC_launcher.sh"
        self.ednaScriptFilePath = os.path.join(self.autoPROCPath, ednaScriptFileName)
        if os.path.exists(self.ednaScriptFilePath):
            # Create unique file name
            ednaScriptFile = tempfile.NamedTemporaryFile(
                suffix=".sh",
                prefix="autoPROC_launcher-",
                dir=self.autoPROCPath,
                delete=False,
            )
            self.ednaScriptFilePath = os.path.join(
                self.autoPROCPath, ednaScriptFile.name
            )
            ednaScriptFile.file.write(self.script)
            ednaScriptFile.close()
        else:
            open(self.ednaScriptFilePath, "w").write(self.script)
        os.chmod(self.ednaScriptFilePath, 0o755)

    def execute(self):
        cmd = (
            "echo 'cd %s;source /mxn/groups/biomax/wmxsoft/scripts_mxcube/biomax_HPC.bash_profile;/mxn/groups/biomax/cmxsoft/edna-mx/scripts_maxiv/edna_sbatch.sh %s' | ssh -F /etc/ssh/.ssh -o UserKnownHostsFile=/etc/ssh/.ssh/known_host -i /etc/ssh/id_rsa_biomax-service %s; source /mxn/groups/biomax/wmxsoft/scripts_mxcube/biomax_HPC.bash_profile"
            % (self.autoPROCPath, self.ednaScriptFilePath, HPC_HOST)
        )

        # for test
        # cmd = "echo 'cd %s;/mxn/groups/biomax/cmxsoft/edna-mx/scripts_maxiv/edna_sbatch.sh %s' | ssh %s" % (autoPROCPath, ednaScriptFilePath, hpc_host)
        # print cmd
        logging.getLogger("HWR").info(
            "Autoproc launcher: command gonna be launched: %s" % cmd
        )

        p = subprocess.Popen(
            cmd, shell=True
        )  # , stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        p.wait()

    def parse_and_execute(self):
        self.parse_input_file()
        self.execute()


if __name__ == "__main__":
    args = sys.argv[1:]

    if (len(args) % 2) != 0:
        logging.error(
            "the argument list is not well formed (odd number of args/options)"
        )
        sys.exit()

    # do the arg parsing by hand since neither getopt nor optparse support
    # single dash long options.
    options = dict()
    for x in range(0, len(args), 2):
        options[args[x]] = args[x + 1]

    autoprocessingPath = options["-path"]

    residues = float(options.get("-residues", 200))
    anomalous = options.get("-anomalous", False)
    spacegroup = options.get("-sg", None)
    cell = options.get("-cell", "0,0,0,0,0,0")
    dataCollectionId = options["-datacollectionID"]
    mode = options.get("-mode")

    autoProc = AutoProcLauncher(
        autoprocessingPath,
        mode,
        dataCollectionId,
        residues,
        anomalous,
        cell,
        spacegroup,
    )
    autoProc.parse_and_execute()



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import copy

import jsonschema

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class DataObject(dict):
    """
    Object intended as a base class for data models
    """

    # See https://www.python.org/dev/peps/pep-0351/
    VERBOSE = True
    _SCHEMA = {}

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.validate()
        self._intset("_mutations", [])
        self._intset("_original", copy.deepcopy(dict(self)))
        self._intset("_previous", copy.deepcopy(dict(self)))

    def _immutable(self, *args, **kwargs):
        raise TypeError(
            "object is immutable, trying to set: %s to %s" % (args[0], args[1])
        )

    _intset = dict.__setattr__
    _setitem = dict.__setitem__
    clear = _immutable
    update = _immutable
    setdefault = _immutable
    pop = _immutable
    popitem = _immutable
    __setitem__ = _immutable
    __setattr__ = _immutable
    __delitem__ = _immutable
    __getattr__ = dict.__getitem__

    def __hash__(self):
        return id(self)

    def dangerously_set(self, key, value):
        """
        Sets the attribute name <key> to value

        Args:
            key (str): Atttribute name
            value (any): The value to set

        Returns:
            None
        """
        if self.VERBOSE:
            print(
                "ImmutableDict (%s), dangerously set %s to %s:"
                % (str(self), key, value)
            )

        self._setitem(key, value)

        try:
            self.validate()
        except jsonschema.exceptions.ValidationError:
            self._setitem(key, self._previous[key])
            raise
        else:
            self._mutations.append((key, value))
            self._intset("_previous", copy.deepcopy(dict(self)))

    def validate(self):
        """
        Validates the attributes against the schema defined in _SCHEMA
        """
        if self._SCHEMA:
            jsonschema.validate(instance=self, schema=self._SCHEMA)

    def to_mutable(self):
        """
        Returns:
            A new standard python dict with all the values
        """
        return copy.deepcopy(dict(self))



import gevent

from mxcubecore.HardwareObjects.abstract.AbstractProcedure import AbstractProcedure
from mxcubecore.model import procedure_model as datamodel


class ProcedureMockup(AbstractProcedure):
    _ARGS_CLASS = (datamodel.MockDataModel,)

    def __init__(self, name):
        super(ProcedureMockup, self).__init__(name)

    def _execute(self, data_model):
        print("Procedure will sleep for %d" % data_model.exposure_time)
        gevent.sleep(data_model.exposure_time)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
import logging
from enum import (
    IntEnum,
    unique,
)

import gevent.event

# Using jsonschma for validating the JSCONSchemas
# https://json-schema.org/
# https://github.com/Julian/jsonschema
from jsonschema import (
    ValidationError,
    validate,
)

from mxcubecore.BaseHardwareObjects import ConfiguredObject
from mxcubecore.dispatcher import dispatcher

# import mxcubecore.model.procedure_model


__credits__ = ["MXCuBE collaboration"]


# Temporary definition should use common denfinition from
# HardwareObject
@unique
class ProcedureState(IntEnum):
    """
    Defines the valid Procedure states
    """

    ERROR = 0
    BUSY = 1
    READY = 3


class AbstractProcedure(ConfiguredObject):
    _ARGS_CLASS = ()
    _KWARGS_CLASS = {}
    _RESULT_CLASS = ()

    @staticmethod
    def set_args_class(args_class, kwargs_class):
        """
        Sets the types of the data models used as arguments, cane be used to
        set the argument classes runtime if the models are built dynamically,
        i.e based on configuration not known before

        Args:
            args_class (tuple[BaseModel]) tuple of classes for args
            kwargs_class (dict[[str]: [BaseModel]]) dictionary containing BaseModels
        """
        AbstractProcedure._ARGS_CLASS = args_class
        AbstractProcedure._KWARGS_CLASS = kwargs_class

    @staticmethod
    def set_result_class(_result_class):
        """
        Sets the types of the data models returned by the Procedure, can
        be used to set the result model runtime of the data model is built
        dynamically, i.e based on configuration not known before

        Returns:
            (tuple[BaseModel]) tuple of classes for args
        """
        AbstractProcedure._RESULT_CLASS = _result_class

    def __init__(self, name):
        super(AbstractProcedure, self).__init__(name)
        self._msg = None
        self._results = None
        self._ready_event = gevent.event.Event()
        self._task = None
        self._state = ProcedureState.READY

        # YML configuration options
        # Category that the Procedure belongs to, configurable through
        # YAML file and used by for listing and displaying the procedure
        # in the right context.
        self.category = ""

    def _init(self):
        pass

    def init(self):
        pass

    def _execute(self, data_model):
        """
        Override to implement main task logic

        Args:
            data_model: sub class of mxcubecore.model.procedure_model
            dict in Python 2.7 and Data class in Python 3.7. Data is validated
            by the data_model object

        Returns:
        """
        pass

    def _pre_execute(self, data_model):
        """
        Override to implement pre execute task logic

        Args:
            data_model: sub class of mxcubecore.model.procedure_model
            Data is validated by the data_model object

        Returns:
        """
        pass

    def _post_execute(self, data_model):
        """
        Override to implement post execute task logic

        Args:
            data_model: sub class of mxcubecore.model.procedure_model
            Data is validated by the data_model object

        Returns:
        """
        pass

    def _set_started(self):
        """
        Emits procedureStarted signal
        Returns:

        """
        self._state = ProcedureState.BUSY
        dispatcher.send(self, "procedureStarted")

    def _set_successful(self):
        """
        Emits procedureSuccessful signal
        Returns:

        """
        self._state = ProcedureState.READY
        dispatcher.send(self, "procedureSuccessful", self.results)

    def _set_error(self):
        """
        Emits procedure error signal
        Returns:

        """
        self._state = ProcedureState.ERROR
        dispatcher.send(self, "procedureError", self.msg)

    def _set_stopped(self):
        """
        Emits procedureStoped signal
        Returns:

        """
        self._state = ProcedureState.READY
        dispatcher.send(self, "procedureStopped", self.results)

    def _start(self, data_model):
        """
        Internal start, for the moment executed in greenlet
        """
        try:
            self._set_started()
            self._pre_execute(data_model)
            self._execute(data_model)
        except Exception as ex:
            self._state = ProcedureState.ERROR
            self._msg = "Procedure execution error (%s)" % str(ex)
            logging.getLogger("HWR").exception(self._msg)
        finally:
            try:
                self._post_execute(data_model)
            except Exception as ex:
                self._state = ProcedureState.ERROR
                self._msg = "Procedure post_execute error (%s)" % str(ex)
                logging.getLogger("HWR").exception(self._msg)

            self._ready_event.set()

            if self._state == ProcedureState.ERROR:
                self._set_error()
            else:
                self._set_successful()

    @property
    def argument_schema(self):
        """
        Schema for arguments passed to start
        Returns:
            dict{"args": tuple[JSONSchema], "kwargs": key: [JSONSchema]}
        """
        return {
            "args": tuple([s.schema_json() for s in self._ARGS_CLASS]),
            "kwargs": {
                key: value.schema_json() for (key, value) in self._KWARGS_CLASS.items()
            },
        }

    @property
    def result_schema(self):
        """
        Schema for result
        Returns:
            tuple[JSONSchema]
        """
        return (s.schema_json() for s in self._RESULT_CLASS)

    @property
    def msg(self):
        """
        Last message produced by procedure
        Returns:
            str
        """
        return self._msg

    @property
    def state(self):
        """
        Execution state
        Returns:
             ProcedureState: The current state of the procedure
        """
        return self._state

    @property
    def results(self):
        """
        Results from procedure execution validated by RESULT_SCHEMA
        if it is defined

        Returns:
            DataClass or frozendict
        """

        return self._results

    def start(self, data_model):
        """
        Starts procedure
        Args:
            data_model: sub class of mxcubecore.model.procedure_model.
            Data is validated by the data_model object

        Returns:
            (Greenlet) The gevent task
        """
        if self._state != ProcedureState.READY:
            self._msg = "Procedure (%s) is already running" % str(self)
            logging.getLogger("HWR").error(self._msg)
        else:
            self._task = gevent.spawn(self._start, data_model)

        return self._task

    def stop(self):
        """
        Stops the execution of procedure
        Returns:
            None
        """
        gevent.kill(self._task)
        self._set_stopped()

    def wait(self):
        """
        Waits for procedure to finish execution

        Returns:
            None
        """
        self._ready_event.wait()



import time

from gevent import Timeout

from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor


class MD2TimeoutError(Exception):
    pass


"""
Example xml file:
<object class="MicrodiffMotor">
  <username>phiy</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <actuator_name>AlignmentY</actuator_name>
  <GUIstep>1.0</GUIstep>
  <unit>-1e-3</unit>
  <resolution>1e-2</resolution>
</object>
"""


class MicrodiffMotor(AbstractMotor):
    (NOTINITIALIZED, UNUSABLE, READY, MOVESTARTED, MOVING, ONLIMIT) = (0, 1, 2, 3, 4, 5)
    EXPORTER_TO_MOTOR_STATE = {
        "Invalid": NOTINITIALIZED,
        "Fault": UNUSABLE,
        "Ready": READY,
        "Moving": MOVING,
        "Created": NOTINITIALIZED,
        "Initializing": NOTINITIALIZED,
        "Unknown": UNUSABLE,
        "Offline": UNUSABLE,
        "LowLim": ONLIMIT,
        "HighLim": ONLIMIT,
    }

    TANGO_TO_MOTOR_STATE = {"STANDBY": READY, "MOVING": MOVING}

    def __init__(self, name):
        AbstractMotor.__init__(self, name)
        self.motor_pos_attr_suffix = "Position"
        self.motor_state_attr_suffix = "State"
        self.translate_state = {
            MicrodiffMotor.NOTINITIALIZED: self.motor_states.NOTINITIALIZED,
            MicrodiffMotor.UNUSABLE: self.motor_states.BUSY,
            MicrodiffMotor.READY: self.motor_states.READY,
            MicrodiffMotor.MOVESTARTED: self.motor_states.MOVESTARTED,
            MicrodiffMotor.MOVING: self.motor_states.MOVING,
            MicrodiffMotor.ONLIMIT: self.motor_states.HIGHLIMIT,
        }

    def init(self):
        self.position = None
        # assign value to actuator_name
        self.actuator_name = self.get_property("actuator_name", "")

        self.GUIstep = self.get_property("GUIstep")

        self.motor_resolution = self.get_property("resolution")
        if self.motor_resolution is None:
            self.motor_resolution = 0.0001

        # this is ugly : I added it to make the centring procedure happy
        self.specName = self.actuator_name

        self.motorState = MicrodiffMotor.NOTINITIALIZED

        self.position_attr = self.get_channel_object(
            "%s%s" % (self.actuator_name, self.motor_pos_attr_suffix)
        )
        if not self.position_attr:
            self.position_attr = self.add_channel(
                {"type": "exporter", "name": "%sPosition" % self.actuator_name},
                self.actuator_name + self.motor_pos_attr_suffix,
            )

        if self.position_attr is not None:
            self.state_attr = self.get_channel_object(
                "%s%s" % (self.actuator_name, self.motor_state_attr_suffix)
            )
            if not self.state_attr:
                self.state_attr = self.add_channel(
                    {"type": "exporter", "name": "%sState" % self.actuator_name},
                    self.actuator_name + self.motor_state_attr_suffix,
                )

            self.position_attr.connect_signal("update", self.motor_positions_changed)
            self.state_attr.connect_signal("update", self.motorStateChanged)

            self.motors_state_attr = self.get_channel_object("motor_states")
            if not self.motors_state_attr:
                self.motors_state_attr = self.add_channel(
                    {"type": "exporter", "name": "motor_states"}, "MotorStates"
                )
            self.motors_state_attr.connect_signal("update", self.updateMotorState)

            self._motor_abort = self.get_command_object("abort")
            if not self._motor_abort:
                self._motor_abort = self.add_command(
                    {"type": "exporter", "name": "abort"}, "abort"
                )

            self.get_dynamic_limits_cmd = self.get_command_object(
                "get%sDynamicLimits" % self.actuator_name
            )
            if not self.get_dynamic_limits_cmd:
                self.get_dynamic_limits_cmd = self.add_command(
                    {
                        "type": "exporter",
                        "name": "get%sDynamicLimits" % self.actuator_name,
                    },
                    "getMotorDynamicLimits",
                )

            self.get_limits_cmd = self.get_command_object("getMotorLimits")
            if not self.get_limits_cmd:
                self.get_limits_cmd = self.add_command(
                    {"type": "exporter", "name": "get_limits"}, "getMotorLimits"
                )

            self.get_max_speed_cmd = self.get_command_object("getMotorMaxSpeed")
            if not self.get_max_speed_cmd:
                self.get_max_speed_cmd = self.add_command(
                    {"type": "exporter", "name": "get_max_speed"}, "getMotorMaxSpeed"
                )

            self.home_cmd = self.get_command_object("homing")
            if not self.home_cmd:
                self.home_cmd = self.add_command(
                    {"type": "exporter", "name": "homing"}, "startHomingMotor"
                )

        self.motor_positions_changed(self.position_attr.get_value())

    def connect_notify(self, signal):
        if signal == "valueChanged":
            self.emit("valueChanged", (self.get_value(),))
        elif signal == "stateChanged":
            self.motorStateChanged(self.state_attr.get_value())
        elif signal == "limitsChanged":
            self.motorLimitsChanged()

    def updateState(self):
        self.set_is_ready(self._get_state() > MicrodiffMotor.UNUSABLE)

    def set_is_ready(self, value):
        if value is True:
            self.set_ready()

    def updateMotorState(self, motor_states):
        d = dict([x.split("=") for x in motor_states])
        # Some are like motors but have no state
        # we set them to ready
        _motor_state = d.get(self.actuator_name)
        if _motor_state is None:
            new_motor_state = MicrodiffMotor.READY
        else:
            if _motor_state in MicrodiffMotor.EXPORTER_TO_MOTOR_STATE:
                new_motor_state = MicrodiffMotor.EXPORTER_TO_MOTOR_STATE[_motor_state]
            else:
                new_motor_state = MicrodiffMotor.TANGO_TO_MOTOR_STATE[_motor_state]
        if self.motorState == new_motor_state:
            return
        self.motorState = new_motor_state
        self.motorStateChanged(self.motorState)

    def motorStateChanged(self, state):
        self.updateState()
        if not isinstance(state, int):
            state = self.get_state()
        self.emit("stateChanged", (state,))

    def _get_state(self):
        state_value = self.state_attr.get_value()
        if state_value in MicrodiffMotor.EXPORTER_TO_MOTOR_STATE:
            self.motorState = MicrodiffMotor.EXPORTER_TO_MOTOR_STATE[state_value]
        else:
            self.motorState = MicrodiffMotor.TANGO_TO_MOTOR_STATE[state_value.name]
        return self.motorState

    def get_state(self):
        state = self._get_state()
        return self.translate_state[state]

    def motorLimitsChanged(self):
        self.emit("limitsChanged", (self.get_limits(),))

    def get_limits(self):
        dynamic_limits = self.get_dynamic_limits()
        if dynamic_limits != (-1e4, 1e4):
            return dynamic_limits
        else:
            try:
                low_lim, hi_lim = map(float, self.get_limits_cmd(self.actuator_name))
                if low_lim == float(1e999) or hi_lim == float(1e999):
                    raise ValueError
                return low_lim, hi_lim
            except Exception:
                return (-1e4, 1e4)

    def get_dynamic_limits(self):
        try:
            low_lim, hi_lim = map(
                float, self.get_dynamic_limits_cmd(self.actuator_name)
            )
            if low_lim == float(1e999) or hi_lim == float(1e999):
                raise ValueError
            return low_lim, hi_lim
        except Exception:
            return (-1e4, 1e4)

    def get_max_speed(self):
        return self.get_max_speed_cmd(self.actuator_name)

    def motor_positions_changed(self, absolute_position, private={}):
        if None not in (absolute_position, self.position):
            if abs(absolute_position - self.position) <= self.motor_resolution:
                return
        self.position = absolute_position
        self.emit("valueChanged", (self.position,))

    def get_value(self):
        if self.position_attr is not None:
            self.position = self.position_attr.get_value()
        return self.position

    def _set_value(self, value):
        # NB these checks are only in update_value
        # If you set the value, you must get the vaue you set
        # if abs(self.position - absolutePosition) >= self.motor_resolution:
        self.position_attr.set_value(value)

    def waitEndOfMove(self, timeout=None):
        with Timeout(timeout):
            time.sleep(0.1)
            while self.motorState == MicrodiffMotor.MOVING:
                time.sleep(0.1)

    def motorIsMoving(self):
        return self.is_ready() and self.motorState == MicrodiffMotor.MOVING

    def get_motor_mnemonic(self):
        return self.actuator_name

    def stop(self):
        if self.get_state() != MicrodiffMotor.NOTINITIALIZED:
            self._motor_abort()

    def homeMotor(self, timeout=None):
        self.home_cmd(self.actuator_name)
        try:
            self.waitEndOfMove(timeout)
        except Exception:
            raise MD2TimeoutError
